#include <iostream>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/connected_components.hpp>
#include <boost/graph/filtered_graph.hpp>

namespace boost {
  enum vertex_component_t { vertex_component = 111 };
  BOOST_INSTALL_PROPERTY(vertex, component);
}

namespace boost {
  enum edge_irreducible_t { edge_irreducible = 112 };
  BOOST_INSTALL_PROPERTY(edge, irreducible);
}

using namespace boost;

template <typename ComponentMap>
struct vertexComponent {

  vertexComponent() {}

  vertexComponent(ComponentMap component, int f_component) : m_component(component), m_f_component(f_component) {}

  template <typename Vertex>
  bool operator()(const Vertex& v) const {
    return (get(m_component, v) == m_f_component);
  }

  ComponentMap m_component;
  int m_f_component;
};

template <typename IrreducibleMap>
struct edgeIrreducible {

  edgeIrreducible() {}

  edgeIrreducible(IrreducibleMap irreducible) : m_irreducible(irreducible) {}

  template <typename Edge>
  bool operator()(const Edge& e) const {
    return get(m_irreducible, e);
  }

  IrreducibleMap m_irreducible;
};

int main(int argc, char ** argv) {

  typedef adjacency_list<vecS, vecS, undirectedS,
                         property<vertex_component_t, int>,
                         property<edge_irreducible_t, bool> > Graph;

  typedef property_map<Graph, vertex_component_t>::type ComponentMap;
  typedef property_map<Graph, edge_irreducible_t>::type IrreducibleMap;
  typedef filtered_graph<Graph,
                         keep_all,
                         vertexComponent<ComponentMap> > FilteredGraph;
  
  typedef filtered_graph<Graph,
                         edgeIrreducible<IrreducibleMap>,
                         vertexComponent<ComponentMap> > ReducedGraph;

  std::pair < graph_traits < Graph >::edge_descriptor, bool > edgePair;
  std::pair < graph_traits < Graph >::edge_descriptor, bool > edgePair2;
  std::pair < graph_traits < Graph >::edge_descriptor, bool > edgePair3;
  graph_traits < Graph >::vertex_iterator ui, ui_end;
  graph_traits < Graph >::vertex_iterator vi, vi_end;
  graph_traits < Graph >::vertex_iterator wi, wi_end;
  graph_traits < Graph >::out_edge_iterator oei, oei_end;
  graph_traits < FilteredGraph >::vertex_iterator fvi, fvi_end;
  graph_traits < FilteredGraph >::out_edge_iterator foei, foei_end;
  graph_traits < ReducedGraph >::vertex_iterator rvi, rvi_end;
  graph_traits < ReducedGraph >::out_edge_iterator roei, roei_end;

  Graph * g;
  FilteredGraph * fg;
  ReducedGraph *rg;

  enum { A, B, C, D, E, F, G, H, I, N };
  const char* name = "ABCDEFGHI";
  g = new Graph(N);
  add_edge(A, B, *g);
  add_edge(C, D, *g);
  add_edge(D, E, *g);
  add_edge(E, C, *g);
  add_edge(F, A, *g);
  add_edge(F, B, *g);
  add_edge(G, H, *g);
  add_edge(G, I, *g);
  add_edge(H, I, *g);

  std::cout<<std::endl<<"Graph out-edges:"<<std::endl;
  for (tie(vi, vi_end) = vertices(*g); vi != vi_end; ++vi) {
    std::cout<<name[*vi]<<" outedges - ";
    for(tie(oei, oei_end)=out_edges(*vi, *g); oei != oei_end; ++oei) {
      std::cout<<name[source(*oei, *g)]<<"-->"<<name[target(*oei, *g)]<<"  ";
    }
    std::cout<<std::endl;
  }
  std::cout<<std::endl;

  int numComponents = connected_components(*g, get(vertex_component, *g));
					   
  std::cout<<"Graph has "<<numComponents<<" components."<<std::endl<<std::endl;

  for(int i=0; i<numComponents; i++) {
    keep_all efilter;
    vertexComponent<ComponentMap> vfilter(get(vertex_component, *g), i);
    fg = new FilteredGraph(*g, efilter, vfilter);

    std::cout<<"Filtered graph (component "<<i<<") out-edges:"<<std::endl;
    for (tie(fvi, fvi_end) = vertices(*fg); fvi != fvi_end; ++fvi) {    
      std::cout<<name[*fvi]<<" outedges - ";    
      for(tie(foei, foei_end)=out_edges(*fvi, *fg); foei != foei_end; ++foei) {
	std::cout<<name[source(*foei, *fg)]<<"-->"<<name[target(*foei, *fg)]<<"  ";
      }
      std::cout<<std::endl;
    }
    std::cout<<std::endl;

    delete fg;
    fg = NULL;
  }

  /* set reducible edges - by eye for now
  edgePair = edge(A, B, *g);
  if(edgePair.second) put(edge_irreducible, *g, edgePair.first, true);
  else { std::cerr<<"bug\n"; return 1; }
  edgePair = edge(B, F, *g);
  if(edgePair.second) put(edge_irreducible, *g, edgePair.first, true);
  else { std::cerr<<"bug\n"; return 1; }
  edgePair = edge(C, D, *g);
  if(edgePair.second) put(edge_irreducible, *g, edgePair.first, true);
  else { std::cerr<<"bug\n"; return 1; }
  edgePair = edge(D, E, *g);
  if(edgePair.second) put(edge_irreducible, *g, edgePair.first, true);
  else { std::cerr<<"bug\n"; return 1; }
  edgePair = edge(G, H, *g);
  if(edgePair.second) put(edge_irreducible, *g, edgePair.first, true);
  else { std::cerr<<"bug\n"; return 1; }
  edgePair = edge(H, I, *g);
  if(edgePair.second) put(edge_irreducible, *g, edgePair.first, true);
  else { std::cerr<<"bug\n"; return 1; }
  */

  // set reducible edges - naive algorithm
  for(tie(ui, ui_end) = vertices(*g); ui != ui_end; ++ui) {
    for(tie(vi, vi_end) = vertices(*g); vi != vi_end; ++vi) {
      for(tie(wi, wi_end) = vertices(*g); wi != wi_end; ++wi) {
	edgePair = edge(*ui, *vi, *g);
	edgePair2 = edge(*vi, *wi, *g);
	edgePair3 = edge(*ui, *wi, *g);
	if(edgePair.second && edgePair2.second) {
	  put(edge_irreducible, *g, edgePair.first, true);
	  put(edge_irreducible, *g, edgePair2.first, true);
	  if(edgePair3.second) put(edge_irreducible, *g, edgePair3.first, false);
	}
      }
    }
  }
  //

  // linear time algorithm

  //

  for(int i=0; i<numComponents; i++) {
    edgeIrreducible<IrreducibleMap> efilter(get(edge_irreducible, *g));
    vertexComponent<ComponentMap> vfilter(get(vertex_component, *g), i);
    rg = new ReducedGraph(*g, efilter, vfilter);

    std::cout<<"Reduced graph (component "<<i<<") out-edges:"<<std::endl;
    for (tie(rvi, rvi_end) = vertices(*rg); rvi != rvi_end; ++rvi) {    
      std::cout<<name[*rvi]<<" outedges - ";    
      for(tie(roei, roei_end)=out_edges(*rvi, *rg); roei != roei_end; ++roei) {
	std::cout<<name[source(*roei, *rg)]<<"-->"<<name[target(*roei, *rg)]<<"  ";
      }
      std::cout<<std::endl;
    }
    std::cout<<std::endl;

    delete rg;
    rg = NULL;
  }

  delete g;
  g = NULL;

  return EXIT_SUCCESS;
}
